use std::{collections::HashMap, process::Stdio, time::Instant};

use htmxpress::Element;
use serde::{Deserialize, Serialize};
use tracing::{debug, warn};

use crate::error::NtitledError;

#[derive(Debug, Deserialize)]
pub struct FFProbeOutput {
    pub streams: Option<Vec<Stream>>,
}

impl FFProbeOutput {
    pub fn probe(path: &str) -> Result<Self, NtitledError> {
        debug!("Probing {path}");

        let start = Instant::now();

        let child = std::process::Command::new("ffprobe")
            .args([
                "-loglevel",
                "error",
                "-show_entries",
                "stream:stream_tags=language",
                "-show_streams",
                "-of",
                "json",
                path,
            ])
            .stdout(Stdio::piped())
            .stderr(Stdio::piped())
            .spawn()?
            .wait_with_output()?;

        debug!(
            "({}ms) Probe complete {}",
            Instant::now().duration_since(start).as_millis(),
            path,
        );

        let out = String::from_utf8(child.stdout)?;
        let err = String::from_utf8(child.stderr)?;
        let err = err.trim();

        if !err.is_empty() {
            warn!("{err}")
        }

        serde_json::from_str::<Self>(&out).map_err(NtitledError::from)
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum Stream {
    Video(VideoStream),
    Audio(AudioStream),
    Sub(SubtitleStream),
}

#[derive(Debug, Serialize, Deserialize)]
pub struct VideoStream {
    pub index: usize,

    #[serde(rename = "codec_name")]
    pub codec: String,

    pub width: u16,

    pub height: u16,

    #[serde(with = "bool_int")]
    pub closed_captions: bool,

    pub avg_frame_rate: String,

    #[serde(rename = "codec_long_name")]
    pub codec_long: String,
    pub profile: Option<String>,
    pub codec_type: String,
    pub coded_width: u16,
    pub coded_height: u16,

    pub start_time: String,
    pub disposition: Disposition,
    pub tags: Option<HashMap<String, String>>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct AudioStream {
    pub index: usize,

    #[serde(rename = "codec_name")]
    pub codec: String,
    #[serde(rename = "codec_long_name")]
    pub codec_long: String,

    pub sample_rate: String,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct SubtitleStream {
    pub index: usize,

    #[serde(rename = "codec_name")]
    pub codec: String,

    pub duration: Option<String>,

    pub tags: Tags,

    #[serde(rename = "codec_long_name")]
    pub codec_long: String,

    pub codec_type: String,
    pub avg_frame_rate: String,
    pub time_base: String,
    pub start_time: String,
    pub duration_ts: Option<usize>,
    pub disposition: Disposition,
}

#[derive(Debug, Serialize, Deserialize, Element)]
pub struct Tags {
    #[element("p")]
    #[format("Language: {}")]
    #[default("unspecified")]
    language: Option<String>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Disposition {
    #[serde(with = "bool_int")]
    pub default: bool,

    #[serde(with = "bool_int")]
    pub dub: bool,

    #[serde(with = "bool_int")]
    pub original: bool,

    #[serde(with = "bool_int")]
    pub comment: bool,
}

mod bool_int {
    use serde::{de::Unexpected, Deserialize, Deserializer, Serializer};

    pub fn serialize<S>(value: &bool, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        match value {
            true => serializer.serialize_u8(1),
            false => serializer.serialize_u8(0),
        }
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<bool, D::Error>
    where
        D: Deserializer<'de>,
    {
        match u8::deserialize(deserializer)? {
            0 => Ok(false),
            1 => Ok(true),
            other => Err(serde::de::Error::invalid_value(
                Unexpected::Unsigned(other as u64),
                &"zero or one",
            )),
        }
    }
}
