use htmxpress::Element;

use crate::{
    ffmpeg::{SubtitleStream, Tags, VideoStream},
    ost::data::{SubtitleFile, SubtitleResponse},
    Directory, File,
};

#[derive(Debug, Element)]
#[element("a")]
#[attrs(class = "file-entry dir")]
#[attr("href" = "/dir/{}", path)]
#[urlencode]
pub struct DirHtmx {
    path: String,

    #[element("h2")]
    #[format("{} &#128193;")]
    name: String,
}

impl DirHtmx {
    pub fn new(dir: Directory) -> Self {
        let Directory { name, path } = dir;
        Self { path, name }
    }
}

#[derive(Debug, Element)]
pub struct DownloadButtonHtmx {
    #[element("button")]
    #[attrs(class = "download-dir-button")]
    #[hx_get("/download/dir?full_path={}", download_path)]
    #[hx("swap" = "innerHtml")]
    #[urlencode]
    download: &'static str,

    download_path: String,
}

impl DownloadButtonHtmx {
    pub fn new(download_path: String) -> Self {
        Self {
            download: "Download",
            download_path,
        }
    }
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "file-entry file")]
pub struct FileHtmx {
    #[nest]
    header: FileHeaderHtmx,

    #[element("div")]
    #[before("<h4>Video</h4>")]
    #[attrs(class = "meta video-meta")]
    #[list(nest)]
    video_meta: Vec<VStreamHtmx>,

    #[element("div")]
    #[before("<h4>Subtitles (Embedded)</h4>")]
    #[attrs(class = "meta sub-meta")]
    #[list(nest)]
    sub_meta: Vec<SStreamHtmx>,

    #[element("div")]
    #[attrs(class = "subtitles")]
    #[nest]
    sub_list: SubtitleListHtmx,
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "file-header")]
struct FileHeaderHtmx {
    id: usize,

    #[element("h2")]
    #[hx_get("/subtitles?name={}&path={}", name, path)]
    #[attr("hx-target"= "#_sub_{}", id)]
    #[map(name => name.rsplit_once('.').map(|n|n.0).unwrap_or(name))]
    #[urlencode]
    name: String,

    path: String,

    /// `true` if a same file exists with a `.srt` extension found
    /// in the same directory or if the file has embedded subs
    #[element("p")]
    #[attrs(class = "has-subs")]
    #[map(el => if *el { "&#10004;" } else { "&#10060;" })]
    has_subs: bool,
}

#[derive(Debug, Element)]
struct SubtitleListHtmx {
    id: usize,

    #[element("div")]
    #[attrs(class = "subtitles-inner")]
    #[attr("id" = "_sub_{}", id)]
    inner: String,
}

impl FileHtmx {
    pub fn new(id: usize, value: File) -> Self {
        let File {
            name,
            path,
            has_subs,
            video_meta,
            sub_meta,
            ..
        } = value;

        Self {
            header: FileHeaderHtmx {
                id,
                name,
                path,
                has_subs,
            },
            video_meta: video_meta.into_iter().map(VStreamHtmx::from).collect(),
            sub_meta: sub_meta.into_iter().map(SStreamHtmx::from).collect(),
            sub_list: SubtitleListHtmx {
                id,
                inner: String::new(),
            },
        }
    }
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "video-meta-inner")]
pub struct VStreamHtmx {
    #[element("p")]
    #[format("Codec: {}")]
    pub codec: String,

    #[element("p")]
    #[format("W: {}")]
    pub width: u16,

    #[element("p")]
    #[format("H: {}")]
    pub height: u16,

    #[element("p")]
    #[format("Captions: {}")]
    pub closed_captions: bool,

    #[element("p")]
    #[format("Framerate: {}")]
    pub avg_frame_rate: String,
}

impl From<VideoStream> for VStreamHtmx {
    fn from(value: VideoStream) -> Self {
        let VideoStream {
            codec,
            width,
            height,
            closed_captions,
            avg_frame_rate,
            ..
        } = value;
        Self {
            codec,
            width,
            height,
            closed_captions,
            avg_frame_rate,
        }
    }
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "sub-meta-inner")]
pub struct SStreamHtmx {
    #[element("p")]
    #[format("Codec: {}")]
    pub codec: String,

    #[element("p")]
    #[format("Duration: {}")]
    #[default("unknown")]
    pub duration: Option<String>,

    #[nest]
    pub tags: Tags,
}

impl From<SubtitleStream> for SStreamHtmx {
    fn from(value: SubtitleStream) -> Self {
        let SubtitleStream {
            codec,
            duration,
            tags,
            ..
        } = value;
        Self {
            codec,
            duration,
            tags,
        }
    }
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "sub-container")]
pub struct SubtitleResponseHtmx {
    #[nest]
    header: SubContainerTitleHtmx,

    #[nest]
    info: SubContainerInfoHtmx,
}

impl SubtitleResponseHtmx {
    pub fn new(path: String, value: SubtitleResponse) -> Self {
        let SubtitleResponse {
            title,
            original_title,
            year,
            download_count,
            parent_title,
            episode_number,
            season_number,
            files,
            language,
            hashmatch,
        } = value;

        let hashmatch = if hashmatch { "🎯" } else { "" };

        let original_title = original_title
            .map(|title| format!("({title}) "))
            .unwrap_or_default();

        let parent_title = parent_title
            .map(|title| format!("{title} -"))
            .unwrap_or_default();

        let episode_id = match (episode_number, season_number) {
            (Some(ep), Some(season)) => format!("S{season}E{ep}"),
            (_, _) => String::new(),
        };

        let files = files
            .into_iter()
            .map(|el| SubFileHtmx::new(path.clone(), el))
            .collect();

        Self {
            header: SubContainerTitleHtmx {
                parent_title,
                title,
                original_title,
                episode_id,
                hashmatch: hashmatch.to_string(),
            },
            info: SubContainerInfoHtmx {
                language,
                downloads: download_count,
                year,
                files,
            },
        }
    }
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "sub-container-title")]
pub struct SubContainerTitleHtmx {
    #[element("h3")]
    #[format("{} {} {}{} {}", title, original_title, episode_id, hashmatch)]
    parent_title: String,
    title: String,
    original_title: String,
    episode_id: String,
    hashmatch: String,
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "sub-container-info")]
pub struct SubContainerInfoHtmx {
    #[element("p")]
    #[attrs(class = "info")]
    #[format("Language: {}")]
    language: Option<String>,

    #[element("p")]
    #[attrs(class = "info")]
    #[format("Downloads: {}")]
    downloads: usize,

    #[element("p")]
    #[attrs(class = "info")]
    #[format("Year: {}")]
    year: usize,

    #[list(nest)]
    files: Vec<SubFileHtmx>,
}

#[derive(Debug, Element)]
#[element("div")]
#[attrs(class = "sub-files")]
pub struct SubFileHtmx {
    file_id: String,
    path: String,

    #[element("button")]
    #[attrs(class = "download-button")]
    #[hx(
        "target" = "closest div",
        "swap" = "innerHtml"
    )]
    #[hx_get("/download?file_id={}&full_path={}", file_id, path)]
    #[urlencode]
    button: &'static str,
}

impl SubFileHtmx {
    fn new(path: String, value: SubtitleFile) -> Self {
        let SubtitleFile { file_id, .. } = value;
        Self {
            file_id: file_id.to_string(),
            path,
            button: "Download",
        }
    }
}
