use axum::{
    response::Redirect,
    routing::{get, get_service},
    Router,
};
use cache::Cache;
use error::NtitledError;
use ffmpeg::VideoStream;
use minijinja::Environment;
use ost::client::OSClient;
use redis::ConnectionInfo;
use serde::{Deserialize, Serialize};
use std::{
    fs::{self, DirEntry},
    num::NonZeroUsize,
    path::Path,
    thread,
};
use tower_http::services::ServeDir;
use tracing::{debug, info, warn, Level};

const SKIP_EXT: &[&str] = &["srt", "smi", "mjpeg", "jpg", "jpeg"];

fn skip_ext(ext: &str) -> bool {
    for e in SKIP_EXT {
        if ext.ends_with(e) {
            return true;
        }
    }
    false
}

use crate::ffmpeg::{FFProbeOutput, SubtitleStream};

pub mod cache;
pub mod error;
pub mod ffmpeg;
pub mod htmx;
pub mod ost;
pub mod routes;

#[derive(Debug, Clone)]
pub struct State {
    client: OSClient,
    env: Environment<'static>,
    base_path: String,
    cache: redis::Client,
}

fn extract_name_ext(entry: &DirEntry) -> Option<(String, String)> {
    let path = entry.path();
    let ext = path.extension()?.to_str()?;
    let (name, _) = path.file_name()?.to_str()?.split_once(&format!(".{ext}"))?;
    Some((name.to_owned(), ext.to_owned()))
}

impl State {
    pub fn scan_dir_contents(
        &mut self,
        path: impl AsRef<Path>,
    ) -> Result<Vec<FileType>, NtitledError> {
        let entries = fs::read_dir(path)?
            .filter_map(Result::ok)
            .collect::<Vec<_>>();

        let subs = entries
            .iter()
            .filter_map(|entry| {
                let (name, ext) = extract_name_ext(entry)?;
                (ext == "srt").then_some(name.to_owned())
            })
            .collect::<Vec<_>>();

        let mut file_list: Vec<FileType> = entries
            .iter()
            .filter_map(|entry| {
                entry
                    .path()
                    .is_dir()
                    .then_some(FileType::Directory(Directory {
                        name: entry.file_name().to_str()?.to_string(),
                        path: entry.path().to_str()?.to_string(),
                    }))
            })
            .collect();

        let mut cached_file_names = vec![];

        for entry in entries.iter() {
            let Some((name, ext)) = extract_name_ext(entry) else {
                continue;
            };

            if skip_ext(&ext) {
                continue;
            }

            let file = self.cache.get_file_meta(&name)?;
            if let Some(file) = file {
                debug!("Found cached file {}", file.name);
                cached_file_names.push(file.name.clone());
                file_list.push(FileType::File(file));
            }
        }

        let scanned_files = thread::scope(|scope| {
            let mut handles = vec![];

            let mut acc = Vec::with_capacity(*MAX_THREADS);

            for _ in 0..*MAX_THREADS {
                acc.push(vec![]);
            }

            let jobs = entries
                .into_iter()
                .enumerate()
                .fold(acc, |mut acc, (i, el)| {
                    acc[i % *MAX_THREADS].push(el);
                    acc
                });

            for entries in jobs {
                let task = scope.spawn(|| {
                    let mut files = vec![];

                    for entry in entries {
                        if entry.path().is_dir() {
                            continue;
                        }

                        let Some((name, ext)) = extract_name_ext(&entry) else {
                            continue;
                        };

                        if skip_ext(&ext) {
                            continue;
                        }

                        if cached_file_names.contains(&name) {
                            continue;
                        }

                        if let Some(file) = Self::extract_file_meta(name, ext, &subs, &entry) {
                            files.push(file)
                        }
                    }

                    files
                });

                handles.push(task);
            }

            let mut files = vec![];

            for handle in handles {
                match handle.join() {
                    Ok(extracted_files) => {
                        files.extend(extracted_files);
                    }
                    Err(e) => {
                        debug!("Error while reading files: {e:?}")
                    }
                }
            }

            files
        });

        for file in scanned_files {
            self.cache.set_file_meta(&file)?;
            file_list.push(FileType::File(file));
        }

        file_list.sort();

        Ok(file_list)
    }

    fn extract_file_meta(
        name: String,
        ext: String,
        subs: &[String],
        entry: &DirEntry,
    ) -> Option<File> {
        let path = entry.path();
        let path_str = path.to_str()?;

        let out = match FFProbeOutput::probe(path_str) {
            Ok(streams) => streams,
            Err(e) => {
                debug!("File {path_str} is not a video: {e}");
                return None;
            }
        };

        let Some(streams) = out.streams else {
            return None;
        };

        let subs_file = subs.contains(&name.to_string());

        let mut video_meta = vec![];
        let mut sub_meta = vec![];

        for stream in streams {
            match stream {
                ffmpeg::Stream::Video(s) => video_meta.push(s),
                ffmpeg::Stream::Sub(s) => sub_meta.push(s),
                ffmpeg::Stream::Audio(s) => {}
            }
        }

        let subs_embedded = !sub_meta.is_empty();

        Some(File {
            name,
            ext,
            path: entry.path().to_str()?.to_string(),
            has_subs: subs_embedded || subs_file,
            video_meta,
            sub_meta,
        })
    }
}

lazy_static::lazy_static! {
    pub static ref INDEX: String =
        std::fs::read_to_string("assets/index.html").expect("missing template");
}

lazy_static::lazy_static! {
    pub static ref MAX_THREADS: usize = std::thread::available_parallelism().unwrap_or(NonZeroUsize::new(1).unwrap()).into();
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt()
        .with_max_level(Level::DEBUG)
        .init();

    info!("Starting ntitled");

    dotenv::dotenv().expect("could not load env");

    info!("Successfully loaded env");

    info!("Max threads: {}", *MAX_THREADS);

    let redis_mode = std::env::var("REDIS_MODE").expect("REDIS_MODE (tcp | unix) not configured");
    let redis_db = std::env::var("REDIS_DB")
        .expect("REDIS_DB not configured")
        .parse()
        .expect("invalid db value");

    let redis_host = std::env::var("REDIS_HOST").expect("REDIS_HOST not configured");

    let conn_info = match redis_mode.as_str() {
        "tcp" => {
            let redis_port = std::env::var("REDIS_PORT")
                .expect("REDIS_PORT not configured")
                .parse()
                .expect("invalid port");

            ConnectionInfo {
                addr: redis::ConnectionAddr::Tcp(redis_host, redis_port),
                redis: redis::RedisConnectionInfo {
                    db: redis_db,
                    username: None,
                    password: None,
                },
            }
        }
        "unix" => ConnectionInfo {
            addr: redis::ConnectionAddr::Unix(redis_host.into()),
            redis: redis::RedisConnectionInfo {
                db: redis_db,
                username: None,
                password: None,
            },
        },
        _ => {
            warn!("invalid REDIS_MODE, default to unix");
            ConnectionInfo {
                addr: redis::ConnectionAddr::Unix(redis_host.into()),
                redis: redis::RedisConnectionInfo {
                    db: redis_db,
                    username: None,
                    password: None,
                },
            }
        }
    };

    let cache = redis::Client::open(conn_info).expect("unable to connect to redis");

    info!("Successfully initialised Redis");

    let client = OSClient::init().await;

    info!("Successfully loaded OST client");

    let base_path = std::env::var("BASE_PATH").expect("base path not configured");
    let port = std::env::var("PORT").unwrap_or(String::from("3001"));

    let mut env = Environment::new();

    env.add_template("index", &INDEX)
        .expect("unable to add index template");

    let state = State {
        client,
        env,
        base_path,
        cache,
    };

    let router = routes::router(state);

    let listener = tokio::net::TcpListener::bind(format!("0.0.0.0:{port}"))
        .await
        .unwrap();

    info!("Starting server on 0.0.0.0:{port}");

    axum::serve(listener, router).await.unwrap();
}

#[derive(Debug, Serialize, PartialEq, Eq, PartialOrd, Ord)]
pub enum FileType {
    Directory(Directory),
    File(File),
}

#[derive(Debug, Serialize, Deserialize)]
pub struct File {
    name: String,
    ext: String,
    path: String,

    /// `true` if a same file exists with a `.srt` extension found
    /// in the same directory or if the file has embedded subs
    has_subs: bool,

    video_meta: Vec<VideoStream>,

    sub_meta: Vec<SubtitleStream>,
}

impl PartialEq for File {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name && self.path == other.path && self.has_subs == other.has_subs
    }
}

impl Eq for File {}

impl PartialOrd for File {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for File {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        match self.name.cmp(&other.name) {
            core::cmp::Ordering::Equal => {}
            ord => return ord,
        }
        match self.path.cmp(&other.path) {
            core::cmp::Ordering::Equal => {}
            ord => return ord,
        }
        self.has_subs.cmp(&other.has_subs)
    }
}

#[derive(Debug, Serialize, PartialEq, Eq, PartialOrd, Ord)]
pub struct Directory {
    name: String,

    path: String,
}
